﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media.Imaging;

namespace TestProject
{
    /// <summary>
    /// 메인 윈도우
    /// </summary>
    public partial class MainWindow : Window
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 회전 벡터
        /// </summary>
        private Vector rotationVector = new Vector();

        /// <summary>
        /// 드래그 시작 포인트
        /// </summary>
        private Point dragStartPoint;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainWindow()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();

            this.viewer.MouseLeftButtonDown += viewer_MouseLeftButtonDown;
            this.viewer.MouseMove           += viewer_MouseMove;
            this.button1.Click              += button_Click;
            this.button2.Click              += button_Click;
            this.button3.Click              += button_Click;
            this.button4.Click              += button_Click;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 뷰어 마우스 왼쪽 버튼 DOWN 처리하기 - viewer_MouseLeftButtonDown(sender, e)

        /// <summary>
        /// 뷰어 마우스 왼쪽 버튼 DOWN 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void viewer_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            dragStartPoint = e.GetPosition(this.viewer);

            this.rotationVector.X = this.viewer.RotationX;
            this.rotationVector.Y = this.viewer.RotationY;
        }

        #endregion
        #region 뷰어 마우스 이동시 처리하기 - viewer_MouseMove(sender, e)

        /// <summary>
        /// 뷰어 마우스 이동시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void viewer_MouseMove(object sender, MouseEventArgs e)
        {
            if(e.LeftButton == MouseButtonState.Released)
            {
                return;
            }

            Vector offsetVector = Point.Subtract(e.GetPosition(this.viewer) , this.dragStartPoint) * 0.25;

            this.viewer.RotationY = this.rotationVector.Y + offsetVector.X;
            this.viewer.RotationX = this.rotationVector.X + offsetVector.Y;
        }

        #endregion
        #region 버튼 클릭시 처리하기 - button_Click(sender, e)

        /// <summary>
        /// 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void button_Click(object sender, RoutedEventArgs e)
        {
            Button button = sender as Button;

            string imageFileName = null;

            switch(button.Name)
            {
                case "button1" : imageFileName = "sample1.jpg"; break;
                case "button2" : imageFileName = "sample2.jpg"; break;
                case "button3" : imageFileName = "sample3.jpg"; break;
                case "button4" : imageFileName = "sample4.jpg"; break;
            }

            Uri uri = new Uri($"pack://application:,,,/{imageFileName}");
            
            BitmapImage bitmapImage = new BitmapImage(uri);

            this.viewer.PanoramaImage = bitmapImage;
        }

        #endregion
    }
}